@@ -1,25 +1,28 @@
-"""Representation of a collection of Doorstop items."""
+"""Representation of a collection of items."""
 
 import os
 from itertools import chain
 from collections import OrderedDict
-import logging
 
-from doorstop.core.base import BaseValidatable
-from doorstop.core.base import auto_load, auto_save, BaseFileObject
-from doorstop.core.types import Prefix, ID, Level
-from doorstop.core.item import Item
 from doorstop import common
 from doorstop.common import DoorstopError, DoorstopWarning
+from doorstop.core.base import (add_document, edit_document, delete_document,
+                                auto_load, auto_save,
+                                BaseValidatable, BaseFileObject)
+from doorstop.core.types import Prefix, UID, Level
+from doorstop.core.item import Item
 from doorstop import settings
 
+log = common.logger(__name__)
 
-class Document(BaseValidatable, BaseFileObject):  # pylint: disable=R0902,R0904
+
+class Document(BaseValidatable, BaseFileObject):  # pylint: disable=R0902
 
     """Represents a document directory containing an outline of items."""
 
     CONFIG = '.doorstop.yml'
     SKIP = '.doorstop.skip'  # indicates this document should be skipped
+    INDEX = 'index.yml'
 
     DEFAULT_PREFIX = Prefix('REQ')
     DEFAULT_SEP = ''
@@ -28,8 +31,8 @@ class Document(BaseValidatable, BaseFileObject):  # pylint: disable=R0902,R0904
     def __init__(self, path, root=os.getcwd(), **kwargs):
         """Initialize a document from an exiting directory.
 
-        @param path: path to document directory
-        @param root: path to root of project
+        :param path: path to document directory
+        :param root: path to root of project
 
         """
         super().__init__()
@@ -55,7 +58,7 @@ class Document(BaseValidatable, BaseFileObject):  # pylint: disable=R0902,R0904
         return "Document('{}')".format(self.path)
 
     def __str__(self):
-        if common.VERBOSITY < common.STR_VERBOSITY:
+        if common.verbosity < common.STR_VERBOSITY:
             return self.prefix
         else:
             return "{} ({})".format(self.prefix, self.relpath)
@@ -63,30 +66,32 @@ class Document(BaseValidatable, BaseFileObject):  # pylint: disable=R0902,R0904
     def __iter__(self):
         yield from self._iter()
 
-    def __eq__(self, other):
-        return isinstance(other, self.__class__) and self.path == other.path
+    def __len__(self):
+        return len(list(self._iter()))
 
-    def __ne__(self, other):
-        return not self == other
+    def __bool__(self):  # override `__len__` behavior, pylint: disable=R0201
+        return True
 
     @staticmethod
+    @add_document
     def new(tree, path, root, prefix, sep=None, digits=None, parent=None, auto=None):  # pylint: disable=R0913,C0301
         """Internal method to create a new document.
 
-        @param tree: reference to tree that contains this document
+        :param tree: reference to tree that contains this document
 
-        @param path: path to directory for the new document
-        @param root: path to root of the project
-        @param prefix: prefix for the new document
+        :param path: path to directory for the new document
+        :param root: path to root of the project
+        :param prefix: prefix for the new document
 
-        @param sep: separator between prefix and numbers
-        @param digits: number of digits for the new document
-        @param parent: parent ID for the new document
-        @param auto: automatically save the document
+        :param sep: separator between prefix and numbers
+        :param digits: number of digits for the new document
+        :param parent: parent UID for the new document
+        :param auto: automatically save the document
 
-        @raise DoorstopError: if the document already exists
+        :raises: :class:`~doorstop.common.DoorstopError` if the document
+            already exists
 
-        @return: new Document
+        :return: new :class:`~doorstop.core.document.Document`
 
         """
         # TODO: raise a specific exception for invalid separator characters?
@@ -96,7 +101,7 @@ class Document(BaseValidatable, BaseFileObject):  # pylint: disable=R0902,R0904
         if os.path.exists(config):
             raise DoorstopError("document already exists: {}".format(path))
         # Create the document directory
-        Document._new(config, name='document')
+        Document._create(config, name='document')
         # Initialize the document
         document = Document(path, root=root, tree=tree, auto=False)
         document.prefix = prefix if prefix is not None else document.prefix
@@ -112,7 +117,7 @@ class Document(BaseValidatable, BaseFileObject):  # pylint: disable=R0902,R0904
         """Load the document's properties from its file."""
         if self._loaded and not reload:
             return
-        logging.debug("loading {}...".format(repr(self)))
+        log.debug("loading {}...".format(repr(self)))
         # Read text from file
         text = self._read(self.config)
         # Parse YAML data from text
@@ -133,9 +138,10 @@ class Document(BaseValidatable, BaseFileObject):  # pylint: disable=R0902,R0904
         if reload:
             list(self._iter(reload=reload))
 
+    @edit_document
     def save(self):
         """Save the document's properties to its file."""
-        logging.debug("saving {}...".format(repr(self)))
+        log.debug("saving {}...".format(repr(self)))
         # Format the data items
         data = {}
         sets = {}
@@ -163,10 +169,11 @@ class Document(BaseValidatable, BaseFileObject):  # pylint: disable=R0902,R0904
     def _iter(self, reload=False):
         """Yield the document's items."""
         if self._itered and not reload:
-            logging.debug("iterating through previously loaded items...")
-            yield from self._items
+            msg = "iterating document {}'s loaded items...".format(self)
+            log.debug(msg)
+            yield from list(self._items)
             return
-        logging.debug("iterating through items in {}...".format(self.path))
+        log.info("loading document {}'s items...".format(self))
         # Reload the document's item
         self._items = []
         for dirpath, dirnames, filenames in os.walk(self.path):
@@ -174,9 +181,8 @@ class Document(BaseValidatable, BaseFileObject):  # pylint: disable=R0902,R0904
                 path = os.path.join(dirpath, dirname, Document.CONFIG)
                 if os.path.exists(path):
                     path = os.path.dirname(path)
-                    msg = "found embedded document: {}".format(path)
-                    logging.debug(msg)
                     dirnames.remove(dirname)
+                    log.trace("skipped embedded document: {}".format(path))
             for filename in filenames:
                 path = os.path.join(dirpath, filename)
                 try:
@@ -186,10 +192,15 @@ class Document(BaseValidatable, BaseFileObject):  # pylint: disable=R0902,R0904
                     pass  # skip non-item files
                 else:
                     self._items.append(item)
+                    if reload:
+                        item.load(reload=reload)
+                    if settings.CACHE_ITEMS and self.tree:
+                        self.tree._item_cache[item.uid] = item  # pylint: disable=W0212
+                        log.trace("cached item: {}".format(item))
         # Set meta attributes
         self._itered = True
         # Yield items
-        yield from self._items
+        yield from list(self._items)
 
     # properties #############################################################
 
@@ -215,14 +226,14 @@ class Document(BaseValidatable, BaseFileObject):  # pylint: disable=R0902,R0904
     @property
     @auto_load
     def sep(self):
-        """Get the prefix-number separator to use for new item IDs."""
+        """Get the prefix-number separator to use for new item UIDs."""
         return self._data['sep']
 
     @sep.setter
     @auto_save
     @auto_load
     def sep(self, value):
-        """Set the prefix-number separator to use for new item IDs."""
+        """Set the prefix-number separator to use for new item UIDs."""
         # TODO: raise a specific exception for invalid separator characters?
         assert not value or value in settings.SEP_CHARS
         self._data['sep'] = value.strip()
@@ -231,14 +242,14 @@ class Document(BaseValidatable, BaseFileObject):  # pylint: disable=R0902,R0904
     @property
     @auto_load
     def digits(self):
-        """Get the number of digits to use for new item IDs."""
+        """Get the number of digits to use for new item UIDs."""
         return self._data['digits']
 
     @digits.setter
     @auto_save
     @auto_load
     def digits(self, value):
-        """Set the number of digits to use for new item IDs."""
+        """Set the number of digits to use for new item UIDs."""
         self._data['digits'] = value
         # TODO: should the new digits be applied to all items?
 
@@ -266,106 +277,239 @@ class Document(BaseValidatable, BaseFileObject):  # pylint: disable=R0902,R0904
         return max(item.depth for item in self)
 
     @property
-    def next(self):
-        """Return the next item number in the document."""
+    def next_number(self):
+        """Get the next item number for the document."""
         try:
-            return max(item.number for item in self) + 1
+            number = max(item.number for item in self) + 1
         except ValueError:
-            return 1
+            number = 1
+        log.debug("next number (local): {}".format(number))
+
+        if self.tree and self.tree.request_next_number:
+            remote_number = 0
+            while remote_number is not None and remote_number < number:
+                if remote_number:
+                    log.warn("server is behind, requesting next number...")
+                remote_number = self.tree.request_next_number(self.prefix)
+                log.debug("next number (remote): {}".format(remote_number))
+            if remote_number:
+                number = remote_number
+
+        return number
 
     @property
     def skip(self):
         """Indicate the document should be skipped."""
         return os.path.isfile(os.path.join(self.path, Document.SKIP))
 
+    @property
+    def index(self):
+        """Get the path to the document's index if it exists else `None`."""
+        path = os.path.join(self.path, Document.INDEX)
+        if os.path.exists(path):
+            return path
+
+    @index.setter
+    def index(self, value):
+        """Create or update the document's index."""
+        if value:
+            path = os.path.join(self.path, Document.INDEX)
+            log.info("creating {} index...".format(self))
+            common.write_lines(self._lines_index(self.items), path)
+
+    @index.deleter
+    def index(self):
+        """Delete the document's index if it exists."""
+        log.info("deleting {} index...".format(self))
+        common.delete(self.index)
+
     # actions ################################################################
 
-    def add_item(self, level=None, reorder=True):
+    # decorators are applied to methods in the associated classes
+    def add_item(self, number=None, level=None, reorder=True):
         """Create a new item for the document and return it.
 
-        @param level: desired item level
-        @param reorder: update levels of document items
+        :param number: desired item number
+        :param level: desired item level
+        :param reorder: update levels of document items
 
-        @return: added Item
+        :return: added :class:`~doorstop.core.item.Item`
 
         """
-        number = self.next
-        logging.debug("next number: {}".format(number))
+        number = max(number or 0, self.next_number)
+        log.debug("next number: {}".format(number))
         try:
             last = self.items[-1]
         except IndexError:
-            nlevel = level
+            next_level = level
         else:
-            nlevel = level or last.level + 1
-        logging.debug("next level: {}".format(nlevel))
-        identifier = ID(self.prefix, self.sep, number, self.digits)
+            if level:
+                next_level = level
+            elif last.level.heading:
+                next_level = last.level >> 1
+                next_level.heading = False
+            else:
+                next_level = last.level + 1
+        log.debug("next level: {}".format(next_level))
+        uid = UID(self.prefix, self.sep, number, self.digits)
         item = Item.new(self.tree, self,
-                        self.path, self.root, identifier,
-                        level=nlevel)
-        self._items.append(item)
+                        self.path, self.root, uid,
+                        level=next_level)
         if level and reorder:
             self.reorder(keep=item)
         return item
 
+    # decorators are applied to methods in the associated classes
     def remove_item(self, value, reorder=True):
-        """Remove an item by its ID.
+        """Remove an item by its UID.
 
-        @param value: item or ID
-        @param reorder: update levels of document items
+        :param value: item or UID
+        :param reorder: update levels of document items
 
-        @raise DoorstopError: if the item cannot be found
+        :raises: :class:`~doorstop.common.DoorstopError` if the item
+            cannot be found
 
-        @return: removed Item
+        :return: removed :class:`~doorstop.core.item.Item`
 
         """
-        identifier = ID(value)
-        item = self.find_item(identifier)
+        uid = UID(value)
+        item = self.find_item(uid)
         item.delete()
-        self._items.remove(item)
         if reorder:
             self.reorder()
         return item
 
-    def reorder(self, items=None, start=None, keep=None):
+    # decorators are applied to methods in the associated classes
+    def reorder(self, manual=True, automatic=True, start=None, keep=None,
+                _items=None):
         """Reorder a document's items.
 
-        @param items: items to reorder (None = reorder instance items)
-        @param start: level to start numbering (None = use current start)
-        @param keep: item or ID to keep over duplicates
+        Two methods are using to create the outline order:
+
+        - manual: specify the order using an updated index file
+        - automatic: shift duplicate levels and compress gaps
+
+        :param manual: enable manual ordering using the index (if one exists)
+
+        :param automatic: enable automatic ordering (after manual ordering)
+        :param start: level to start numbering (None = use current start)
+        :param keep: item or UID to keep over duplicates
 
         """
-        items = items or self.items
-        keep = self.find_item(keep) if keep else None
-        logging.info("reordering {}...".format(self))
-        self._reorder(items, start=start, keep=keep)
+        # Reorder manually
+        if manual and self.index:
+            log.info("reordering {} from index...".format(self))
+            self._reorder_from_index(self, self.index)
+            del self.index
+        # Reorder automatically
+        if automatic:
+            log.info("reordering {} automatically...".format(self))
+            items = _items or self.items
+            keep = self.find_item(keep) if keep else None
+            self._reorder_automatic(items, start=start, keep=keep)
 
     @staticmethod
-    def _reorder(items, start=None, keep=None):
-        """Reorder a document's items.
+    def _lines_index(items):
+        """Generate (pseudo) YAML lines for the document index."""
+        yield '#' * settings.MAX_LINE_LENGTH
+        yield '# THIS TEMPORARY FILE WILL BE DELETED AFTER DOCUMENT REORDERING'
+        yield '# MANUALLY INDENT, DEDENT, & MOVE ITEMS TO THEIR DESIRED LEVEL'
+        yield '# CHANGES ARE BE REFLECTED IN THE ITEM FILES AFTER CONFIRMATION'
+        yield '#' * settings.MAX_LINE_LENGTH
+        yield ''
+        yield "initial: {}".format(items[0].level if items else 1.0)
+        yield "outline:"
+        for item in items:
+            space = "    " * item.depth
+            comment = item.text.replace('\n', ' ') or item.ref
+            line = space + "- {u}: # {c}".format(u=item.uid, c=comment)
+            if len(line) > settings.MAX_LINE_LENGTH:
+                line = line[:settings.MAX_LINE_LENGTH - 3] + '...'
+            yield line
 
-        @param items: items to reorder
-        @param start: level to start numbering (None = use current start)
-        @param keep: item to keep over duplicates
+    @staticmethod
+    def _reorder_from_index(document, path):
+        """Reorder a document's item from the index."""
+        # Load and parse index
+        text = common.read_text(path)
+        data = common.load_yaml(text, path)
+        # Read updated values
+        initial = data.get('initial', 1.0)
+        outline = data.get('outline', [])
+        # Update levels
+        level = Level(initial)
+        Document._reorder_section(outline, level, document)
+
+    @staticmethod
+    def _reorder_section(section, level, document):
+        """Recursive function to reorder a section of an outline.
+
+        :param section: recursive `list` of `dict` loaded from document index
+        :param level: current :class:`~doorstop.core.types.Level`
+        :param document: :class:`~doorstop.core.document.Document` to order
+
+        """
+        if isinstance(section, dict):  # a section
+
+            # Get the item and subsection
+            uid = list(section.keys())[0]
+            try:
+                item = document.find_item(uid)
+            except DoorstopError as exc:
+                log.debug(exc)
+                item = None
+            subsection = section[uid]
+
+            # An item is a header if it has a subsection
+            level.heading = bool(subsection)
+
+            # Apply the new level
+            if item is None:
+                log.info("({}): {}".format(uid, level))
+            elif item.level == level:
+                log.info("{}: {}".format(item, level))
+            else:
+                log.info("{}: {} to {}".format(item, item.level, level))
+            if item:
+                item.level = level
+
+            # Process the heading's subsection
+            if subsection:
+                Document._reorder_section(subsection, level >> 1, document)
+
+        elif isinstance(section, list):  # a list of sections
+
+            # Process each subsection
+            for index, subsection in enumerate(section):
+                Document._reorder_section(subsection, level + index, document)
+
+    @staticmethod
+    def _reorder_automatic(items, start=None, keep=None):
+        """Reorder a document's items automatically.
+
+        :param items: items to reorder
+        :param start: level to start numbering (None = use current start)
+        :param keep: item to keep over duplicates
 
         """
         nlevel = plevel = None
         for clevel, item in Document._items_by_level(items, keep=keep):
-            logging.debug("current level: {}".format(clevel))
+            log.debug("current level: {}".format(clevel))
             # Determine the next level
             if not nlevel:
                 # Use the specified or current starting level
                 nlevel = Level(start) if start else clevel
                 nlevel.heading = clevel.heading
-                logging.debug("next level (start): {}".format(nlevel))
+                log.debug("next level (start): {}".format(nlevel))
             else:
                 # Adjust the next level to be the same depth
                 if len(clevel) > len(nlevel):
                     nlevel >>= len(clevel) - len(nlevel)
-                    logging.debug("matched current indent: {}".format(nlevel))
+                    log.debug("matched current indent: {}".format(nlevel))
                 elif len(clevel) < len(nlevel):
                     nlevel <<= len(nlevel) - len(clevel)
                     # nlevel += 1
-                    logging.debug("matched current dedent: {}".format(nlevel))
+                    log.debug("matched current dedent: {}".format(nlevel))
                 nlevel.heading = clevel.heading
                 # Check for a level jump
                 _size = min(len(clevel.value), len(plevel.value))
@@ -375,22 +519,22 @@ class Document(BaseValidatable, BaseFileObject):  # pylint: disable=R0902,R0904
                         nlevel += 1
                         nlevel >>= len(clevel) - len(nlevel)
                         msg = "next level (jump): {}".format(nlevel)
-                        logging.debug(msg)
+                        log.debug(msg)
                         break
                 # Check for a normal increment
                 else:
                     if len(nlevel) <= len(plevel):
                         nlevel += 1
                         msg = "next level (increment): {}".format(nlevel)
-                        logging.debug(msg)
+                        log.debug(msg)
                     else:
                         msg = "next level (indent/dedent): {}".format(nlevel)
-                        logging.debug(msg)
+                        log.debug(msg)
             # Apply the next level
             if clevel == nlevel:
-                logging.info("{}: {}".format(item, clevel))
+                log.info("{}: {}".format(item, clevel))
             else:
-                logging.info("{}: {} to {}".format(item, clevel, nlevel))
+                log.info("{}: {} to {}".format(item, clevel, nlevel))
             item.level = nlevel.copy()
             # Save the current level as the previous level
             plevel = clevel.copy()
@@ -410,71 +554,75 @@ class Document(BaseValidatable, BaseFileObject):  # pylint: disable=R0902,R0904
             # Reorder items at this level
             if keep in items:
                 # move the kept item to the front of the list
-                logging.debug("keeping {} level over duplicates".format(keep))
+                log.debug("keeping {} level over duplicates".format(keep))
                 items = [items.pop(items.index(keep))] + items
             for item in items:
                 yield level, item
 
     def find_item(self, value, _kind=''):
-        """Return an item by its ID.
+        """Return an item by its UID.
 
-        @param value: item or ID
+        :param value: item or UID
 
-        @raise DoorstopError: if the item cannot be found
+        :raises: :class:`~doorstop.common.DoorstopError` if the item
+            cannot be found
 
-        @return: matching Item
+        :return: matching :class:`~doorstop.core.item.Item`
 
         """
-        identifier = ID(value)
+        uid = UID(value)
         for item in self:
-            if item.id == identifier:
+            if item.uid == uid:
                 return item
 
-        raise DoorstopError("no matching{} ID: {}".format(_kind, identifier))
+        raise DoorstopError("no matching{} UID: {}".format(_kind, uid))
 
     def get_issues(self, item_hook=None, **kwargs):
         """Yield all the document's issues.
 
-        @param item_hook: function to call for custom item validation
+        :param item_hook: function to call for custom item validation
 
-        @return: generator of DoorstopError, DoorstopWarning, DoorstopInfo
+        :return: generator of :class:`~doorstop.common.DoorstopError`,
+                              :class:`~doorstop.common.DoorstopWarning`,
+                              :class:`~doorstop.common.DoorstopInfo`
 
         """
         assert kwargs.get('document_hook') is None
         hook = item_hook if item_hook else lambda **kwargs: []
-        logging.info("checking document {}...".format(self))
+        log.info("checking document {}...".format(self))
         # Check for items
         items = self.items
         if not items:
-            return DoorstopWarning("no items")
+            yield DoorstopWarning("no items")
+            return
         # Reorder or check item levels
         if settings.REORDER:
-            self.reorder(items=items)
+            self.reorder(_items=items)
         elif settings.CHECK_LEVELS:
             yield from self._get_issues_level(items)
         # Check each item
-        for item in self:
+        for item in items:
             # Check item
             for issue in chain(hook(item=item, document=self, tree=self.tree),
                                item.get_issues()):
-                # Prepend the item's ID to yielded exceptions
+                # Prepend the item's UID to yielded exceptions
                 if isinstance(issue, Exception):
-                    yield type(issue)("{}: {}".format(item.id, issue))
+                    yield type(issue)("{}: {}".format(item.uid, issue))
 
     @staticmethod
     def _get_issues_level(items):
         """Yield all the document's issues related to item level."""
         prev = items[0] if items else None
         for item in items[1:]:
-            pid = prev.id
+            puid = prev.uid
             plev = prev.level
-            nid = item.id
+            nuid = item.uid
             nlev = item.level
-            logging.debug("checking level {} to {}...".format(plev, nlev))
+            log.debug("checking level {} to {}...".format(plev, nlev))
             # Duplicate level
             if plev == nlev:
-                ids = sorted((pid, nid))
-                msg = "duplicate level: {} ({}, {})".format(plev, *ids)
+                uids = sorted((puid, nuid))
+                msg = "duplicate level: {} ({}, {})".format(plev, *uids)
                 yield DoorstopWarning(msg)
             # Skipped level
             length = min(len(plev.value), len(nlev.value))
@@ -487,14 +635,15 @@ class Document(BaseValidatable, BaseFileObject):  # pylint: disable=R0902,R0904
                         (plev.value[index] != nlev.value[index] and
                          index + 1 < length and
                          nlev.value[index + 1] not in (0, 1))):
-                    msg = "skipped level: {} ({}), {} ({})".format(plev, pid,
-                                                                   nlev, nid)
+                    msg = "skipped level: {} ({}), {} ({})".format(plev, puid,
+                                                                   nlev, nuid)
                     yield DoorstopWarning(msg)
                     break
             prev = item
 
+    @delete_document
     def delete(self, path=None):
         """Delete the document and its items."""
         for item in self:
             item.delete()
-        super().delete(self.config)
+        # the document is deleted in the decorated method
